/* 
 *$Header: /usr/src/mash/repository/mash/mash-1/codec/encoder-ldct.cc,v 1.5 1997/08/27 06:58:52 mccanne Exp $ 
 */

#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "tclcl.h"

#include "config.h"
#include "net.h"
#include "rtp.h"
#include "crdef.h"
#include "module.h"
#include "pktbuf-rtp.h"
#include "dct.h"

#include "ldct-ht.h"

#define LDCT_NLAYER 	3	/* Number of layers in ldct codec */

struct ldct_layer {
	pktbuf* pb;

	/* bit buffers */
	u_int bb;
	u_int nbb;
	u_char* bs;
	u_char* es;	

  	/* Refinement buffers */
	u_int refbb;
	u_int refnbb;
	u_char* refbs;
	u_char ref[3*256/8];

	/* Huffman table structures */
	const struct huffent* ht;
	const int* esc;

	int qd;			/* Spatial layering state */
	int layer;		/* Network layer number */

	/* XXX */
	ldct_layer* mergep;	/* Merge pointer for temporary layers */
	u_char* buf;		/* Buffer space for temp layers */
};

struct spatial_hierarchy {
	int n;				/* number of layers in time slot */
	ldct_layer ls[LDCT_NLAYER];	/* spatial layers */
};

class LDCTEncoder : public EncoderModule {
 public:
	LDCTEncoder();
	~LDCTEncoder();
	virtual void recv(Buffer*) = 0;
	virtual int command(int argc, const char*const* argv);
 protected:
	void size(int w, int h); 
	void getpkt(ldct_layer* lp, int S);
	void flush(ldct_layer* lp, int sync, int eog);
	void flush(int i, int sync);
	void setup_layers(int i);
	void setup_lps();
	void check_buffers();

	virtual void encode_blk(const u_char* in) = 0;
	virtual void encode_color(const u_char* blk) = 0;
	void finish_comp(void);

	/* XXX */
	ldct_layer tmpls_[LDCT_NLAYER];

	int blkno_;
	int group_;
	int maxlayer_;

	ldct_layer* lp_[LDCT_NLAYER];
	spatial_hierarchy* shs_;
	spatial_hierarchy* sh_;

	float qt_[64];

	void setq();
	void set_lastlayers();

	int cc_;
	u_int32_t ts_;
};

class LDCTPixelEncoder : public LDCTEncoder {
public:
	LDCTPixelEncoder();
	virtual void recv(Buffer*);
private:
	virtual void encode_blk(const u_char* in);
	virtual void encode_color(const u_char* blk);
};

class LDCTDCTEncoder : public LDCTEncoder {
public:
	LDCTDCTEncoder();	
	virtual void recv(Buffer*);
private:
	virtual void encode_blk(const u_char* in);
	virtual void encode_color(const u_char* blk);
};

static class LDCTPixelEncoderClass : public TclClass {
public:
	LDCTPixelEncoderClass() : TclClass("Module/VideoEncoder/Pixel/LDCT") {}
	TclObject* create(int, const char*const*) {
		return (new LDCTPixelEncoder);
	}
} pixel_matcher;

static class LDCTDCTEncoderClass : public TclClass {
public:
	LDCTDCTEncoderClass() : TclClass("Module/VideoEncoder/DCT/LDCT") {}
	TclObject* create(int, const char*const*) {
		return (new LDCTDCTEncoder);
	}
} dct_matcher;

/*XXX Code at this point is written for 32 bits only */
#define NBIT 32

#if BYTE_ORDER == LITTLE_ENDIAN
#define STORE_BITS(bs, bb) \
	((u_char*)bs)[0] = bb >> 24; \
	((u_char*)bs)[1] = bb >> 16; \
	((u_char*)bs)[2] = bb >> 8; \
	((u_char*)bs)[3] = bb;
#else
#define STORE_BITS(bs, bb) *(u_int*)(bs) = (bb);
#endif

#define PUT_BITS(bits, n, nbb, bb, bs) \
{ \
	nbb += (n); \
	if (nbb > NBIT)  { \
		int extra__ = (nbb) - NBIT; \
		bb |= (bits) >> extra__; \
		STORE_BITS(bs, bb) \
		bs += (NBIT / 8); \
		bb = (bits) << NBIT - extra__; \
		nbb = extra__; \
	} else \
		bb |= (bits) << NBIT - (nbb); \
}

/* Optimize common case */
#define PUT_BIT(bit, nbb, bb, bs) \
{ \
	nbb++; \
	if (nbb > NBIT)  { \
		STORE_BITS(bs, bb) \
		bs += (NBIT / 8); \
		bb = (bit) << NBIT - 1; \
		nbb = 1; \
	} else \
		bb |= (bit) << NBIT - (nbb); \
}

/* No quantization on the dct coefficients */
static const int qt[64] = {
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1, 
	1, 1, 1, 1, 1, 1, 1, 1
};

static const struct huffent* ldct_hte[] = {
	hte_ldct_layer0, hte_ldct_layer1, hte_ldct_layer2 };

static const int* ldct_esc[] = {
	hte_ldct_layer0_esc, hte_ldct_layer1_esc, hte_ldct_layer2_esc };

#define HDRSIZE (sizeof(rtphdr) + sizeof(ldcthdr))

LDCTEncoder::LDCTEncoder()
     	    : group_(0), shs_(0), sh_(0), maxlayer_(4)
{
	/* Temporary layer storage */
	memset(tmpls_, 0x0, LDCT_NLAYER * sizeof(ldct_layer));

	int i;
	for (i = 0; i < LDCT_NLAYER; i++) {
		ldct_layer* lp = &tmpls_[i];
		lp->buf = new u_char[2 * RTP_MTU];
		lp->ht = ldct_hte[i];
		lp->esc = ldct_esc[i];
	}

	/* XXX */
	width_ = 0;
	height_ = 0;

	fdct_fold_q(qt, qt_);
	
	setq();
	set_lastlayers();
}

LDCTEncoder::~LDCTEncoder()
{
	delete shs_;
}

LDCTPixelEncoder::LDCTPixelEncoder()
{
}

LDCTDCTEncoder::LDCTDCTEncoder()
{
}

void LDCTEncoder::setq()
{
/*XXX*/
#define NT 4

	delete shs_;
	shs_ = new spatial_hierarchy[NT];

	/* XXX */
	memset(shs_, 0, NT * sizeof(*shs_));
	shs_[0].n = 3;
	shs_[0].ls[0].qd = LDCT_QD_CREATE(0, 0, 0);
	shs_[0].ls[1].qd = LDCT_QD_CREATE(1, 0, 0);
	shs_[0].ls[2].qd = LDCT_QD_CREATE(2, 0, 1);
	shs_[0].ls[0].mergep = shs_[0].ls[1].mergep = shs_[0].ls[2].mergep = 0;
	shs_[0].ls[0].layer = 0;
	shs_[0].ls[1].layer = 1;
	shs_[0].ls[2].layer = 4;
	
	shs_[1].n = 2;
	shs_[1].ls[0].qd = LDCT_QD_CREATE(0, 1, 0);
	shs_[1].ls[1].qd = LDCT_QD_CREATE(2, 0, 1);
	shs_[1].ls[0].mergep = shs_[1].ls[1].mergep = 0;
	shs_[1].ls[0].layer = 3;
	shs_[1].ls[1].layer = 4;

	shs_[2].n = 2;
	shs_[2].ls[0].qd = LDCT_QD_CREATE(0, 1, 0);
	shs_[2].ls[1].qd = LDCT_QD_CREATE(2, 0, 1);
	shs_[2].ls[0].mergep = shs_[2].ls[1].mergep = 0;
	shs_[2].ls[0].layer = 2;
	shs_[2].ls[1].layer = 4;

	shs_[3].n = 2;
	shs_[3].ls[0].qd = LDCT_QD_CREATE(0, 1, 0);
	shs_[3].ls[1].qd = LDCT_QD_CREATE(2, 0, 1);
	shs_[3].ls[0].mergep = shs_[3].ls[1].mergep = 0;
	shs_[3].ls[0].layer = 3;
	shs_[3].ls[1].layer = 4;
}


void LDCTEncoder::set_lastlayers()
{
#define LDCT_LL	(1 << 6)
	int i;
	for (i = 0; i < NT; i++) {
		spatial_hierarchy* p = &shs_[i];
		int n = p->n;
		int last = 0;
		int j;
		for (j = 0; j < n; j++) {
			if (p->ls[j].layer > maxlayer_)
				break;
			p->ls[j].qd &=~ LDCT_LL;
			last = j;
		}
		p->ls[last].qd |= LDCT_LL;
	}
}

void LDCTEncoder::flush(ldct_layer* lp, int sync, int eog)
{
	/* flush bit buffer */
	STORE_BITS(lp->bs, lp->bb);

	pktbuf* pb = lp->pb;
	int nbit = (lp->bs - &pb->data[HDRSIZE] << 3) + lp->nbb;
	int cc = (nbit + 7) >> 3;
	int ebit = (cc << 3) - nbit;

	/*XXX*/
	if (cc == 0 && sync == 0)
		abort();

	pb->len = cc + HDRSIZE;
	rtphdr* rh = (rtphdr*)pb->data;
	ldcthdr* bh = (ldcthdr*)(rh + 1);

	if (sync && (lp - sh_->ls) == 0)
		rh->rh_flags |= htons(RTP_M);

	if (eog)
		/* XXX */
		bh->E = 1;
	bh->ebit = ebit;

	target_->recv(pb);

	lp->pb = 0;
	cc_ += cc + sizeof(rtphdr) + sizeof(ldcthdr);
}

int LDCTEncoder::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "maxlayer") == 0) {
			maxlayer_ = atoi(argv[2]);
			set_lastlayers();
			return (TCL_OK);
		}
	}
	return (EncoderModule::command(argc, argv));
}

extern void loopback(const u_char* frm, int cc);

void LDCTEncoder::getpkt(ldct_layer* lp, int S)
{
	pktbuf* pb = pool_->alloc(ts_, RTP_PT_LDCT, lp->layer);
	if (lp->pb != 0 || lp->layer != pb->layer)
		abort();
	lp->pb = pb;
	rtphdr* rh = (rtphdr*)pb->data;
	ldcthdr* bh = (ldcthdr*)(rh + 1);
	bh->version = 0;
	bh->width = width_ >> 3;
	bh->height = height_ >> 3;
	bh->qd = htons(lp->qd);
	bh->blkno = htons(blkno_);
	bh->group = group_;
	bh->S = S;
	bh->E = 0;
	int nl = (lp - sh_->ls) + 1;
	bh->nl = (nl < sh_->n) ? sh_->ls[nl].layer : 0;
	lp->bs = &pb->data[HDRSIZE];
	lp->es = lp->bs + mtu_ - HDRSIZE;
	lp->bb = 0;
	lp->nbb = 0;
}

extern "C" const u_int ldcttab[];

#define RLEN(r) ((r) & 0x3ff)

#define RLENVEC(o) ((o) & (3 << 20 | 3 << 10 | 3))
#define REF(o) (((o) >> 2) & 0x7)
#define NREF(o) (((o) >> 5) & 0x3)
#define COUT(o) ((o) & (1 << 7))

#define ONELAYER(r, o, nbb, bb, bs, sign, layer, ht, esc, refnbb, refbb, \
		 refbs) \
{ \
	int nref__ = NREF(o); \
	if (COUT(o)) { \
		u_int rlen__ = RLEN(r >> 10 * layer); \
		/* Output the code */ \
		u_int v__ = ht[rlen__].val; \
		int nb__ = ht[rlen__].nb; \
		PUT_BITS(v__, nb__, nbb, bb, bs); \
		if (esc[rlen__]) { \
			PUT_BITS(rlen__, LDCT_ACMAXCODE, nbb, bb, bs); \
		} \
		/* Zero run length for layer */ \
		r &=~ (0x3ff << 10 * layer); \
		if (nref__ != 0) { \
			u_int ref__ = sign << nref__ | REF(o); \
			PUT_BITS(ref__, nref__ + 1, nbb, bb, bs); \
		} else \
			PUT_BIT(sign, nbb, bb, bs); \
	} else if (nref__ != 0) { \
		u_int ref__ = REF(o); \
		PUT_BITS(ref__, nref__, refnbb, refbb, refbs); \
	} \
}

#define FIRSTLAYER(r, o, nbb, bb, bs, sign, ht, esc) \
{ \
	if (COUT(o)) { \
		u_int rlen__ = RLEN(r); \
		/* Output the code */ \
		u_int v__ = ht[rlen__].val; \
		int nb__ = ht[rlen__].nb; \
		PUT_BITS(v__, nb__, nbb, bb, bs); \
		if (esc[rlen__]) { \
			PUT_BITS(rlen__, LDCT_ACMAXCODE, nbb, bb, bs); \
		} \
		/* Zero run length for layer */ \
		r &=~ 0x3ff; \
		int nref__ = NREF(o); \
		if (nref__ != 0) { \
			u_int ref__ = sign << nref__ | REF(o); \
			PUT_BITS(ref__, nref__ + 1, nbb, bb, bs); \
		} else \
			PUT_BIT(sign, nbb, bb, bs); \
	} \
}

#define ONECOEFF(r, b, lp0, lp1, lp2) \
{ \
	if (b != 0) { \
		int sign__; \
		if (b < 0) { \
			b = -b; \
			sign__ = 1; \
		} else { \
			sign__ = 0; \
		} \
		if (b >= (1<<9)) \
			b = (1<<9) - 1; \
 \
		u_int o__ = ldcttab[b >> 2]; \
		r += RLENVEC(o__); \
		if (lp0->layer <= maxlayer_) { \
		       FIRSTLAYER(r, o__, lp0->nbb, lp0->bb, lp0->bs, sign__, \
			   lp0->ht, lp0->esc);\
		} \
		if (lp1->layer <= maxlayer_) { \
	        	o__ >>= 10; \
		        ONELAYER(r, o__, lp1->nbb, lp1->bb, lp1->bs, sign__, \
				 1, lp1->ht, lp1->esc, lp1->refnbb, \
				 lp1->refbb, lp1->refbs); \
		 } \
		if (lp2->layer <= maxlayer_) { \
			o__ >>= 10; \
			ONELAYER(r, o__, lp2->nbb, lp2->bb, lp2->bs, sign__, \
				 2, lp2->ht, lp2->esc, lp2->refnbb, \
				 lp2->refbb, lp2->refbs); \
		 } \
	} else \
		r += RLENVEC(ldcttab[0]); \
}


#define PUT_BUFFER(nbb, bb, bs, buf, bufnb) \
{ \
	register u_char* p__ = buf;\
	u_int bits__;\
	while(bufnb > 8) {\
		bits__ = *p__++; \
		PUT_BITS(bits__, 8, nbb, bb, bs); \
		bufnb -= 8; \
	} \
	bits__ = *p__ >> 8 - bufnb; \
	PUT_BITS(bits__, bufnb, nbb, bb, bs); \
}		

#define DCENCODE(indc, nbb, bb, bs, dcpred) \
{\
	 int dc__ = indc >> 3; \
	 int delta__ = dc__ - dcpred + 256; \
	 int v__ = hte_ldct_dchuff[delta__].val; \
	 int nb__ = hte_ldct_dchuff[delta__].nb; \
	 PUT_BITS(v__, nb__, nbb, bb, bs); \
	 if (hte_ldct_dchuff_esc[delta__]) { \
		   PUT_BITS(delta__, LDCT_DCMAXCODE, nbb, bb, bs); \
	 } \
	 dcpred = dc__; \
}

void LDCTPixelEncoder::encode_blk(const u_char* in)
{
	short blk[256];

	short* b0 = blk;
	short* b1 = b0 + 64;
	short* b2 = b1 + 64;
	short* b3 = b2 + 64;

	const u_char* p = in;
	fdct(p, width_, b0, qt_);
	fdct(p + 8, width_, b1, qt_);
	p += width_ << 3;
	fdct(p, width_, b2, qt_);
	fdct(p + 8, width_, b3, qt_);
	
	register ldct_layer* lp0 = lp_[0];
	register ldct_layer* lp1 = lp_[1];
	register ldct_layer* lp2 = lp_[2];

	int nbb = lp0->nbb;
	u_int bb = lp0->bb;
	u_char* bs = lp0->bs;
	int dcpred = b0[0] >> 3;
	PUT_BITS(dcpred, 8, nbb, bb, bs);
	DCENCODE(b1[0], nbb, bb, bs, dcpred);
	DCENCODE(b2[0], nbb, bb, bs, dcpred);
	DCENCODE(b3[0], nbb, bb, bs, dcpred);
	lp0->nbb = nbb;
	lp0->bb = bb;
	lp0->bs = bs;

	lp1->refbs = lp1->ref;
	lp2->refbs = lp2->ref;
	lp1->refnbb = lp2->refnbb = 0;
	lp1->refbb = lp2->refbb = 0;

	u_int r = 0;
	for (int i = 1; i < 64; i++) {
		int z = COLZAG[i];
		register int b = b0[z];
		ONECOEFF(r, b, lp0, lp1, lp2);
		b = b1[z];
		ONECOEFF(r, b, lp0, lp1, lp2);
		b = b2[z];
		ONECOEFF(r, b, lp0, lp1, lp2);
		b = b3[z];
		ONECOEFF(r, b, lp0, lp1, lp2);
	}
	
	finish_comp();
}

void LDCTPixelEncoder::encode_color(const u_char* in)
{
	short blk[8*8];

	short* b0 = blk;

	fdct(in, width_ >> 1, b0, qt_);

	register ldct_layer* lp0 = lp_[0];
	register ldct_layer* lp1 = lp_[1];
	register ldct_layer* lp2 = lp_[2];

	int dc = b0[0] >> 3;
	PUT_BITS(dc, 8, lp0->nbb, lp0->bb, lp0->bs);

	lp1->refbs = lp1->ref;
	lp2->refbs = lp2->ref;
	lp1->refnbb = lp2->refnbb = 0;
	lp1->refbb = lp2->refbb = 0;

	u_int r = 0;
	for (int i = 1; i < 64; i++) {
		int z = COLZAG[i];
		register int b = b0[z];
		ONECOEFF(r, b, lp0, lp1, lp2);
	}

	finish_comp();
}

void LDCTDCTEncoder::encode_blk(const u_char* in)
{
	const short* b0 = (const short*)in;
	const short* b1 = b0 + 64;
	const short* b2 = b1 + 64;
	const short* b3 = b2 + 64;

	register ldct_layer* lp0 = lp_[0];
	register ldct_layer* lp1 = lp_[1];
	register ldct_layer* lp2 = lp_[2];

	int nbb = lp0->nbb;
	u_int bb = lp0->bb;
	u_char* bs = lp0->bs;
	int dcpred = b0[0] >> 3;
	PUT_BITS(dcpred, 8, nbb, bb, bs);
	DCENCODE(b1[0], nbb, bb, bs, dcpred);
	DCENCODE(b2[0], nbb, bb, bs, dcpred);
	DCENCODE(b3[0], nbb, bb, bs, dcpred);
	lp0->nbb = nbb;
	lp0->bb = bb;
	lp0->bs = bs;

	lp1->refbs = lp1->ref;
	lp2->refbs = lp2->ref;
	lp1->refnbb = lp2->refnbb = 0;
	lp1->refbb = lp2->refbb = 0;

	u_int r = 0;
	for (int i = 1; i < 64; i++) {
		int z = COLZAG[i];
		register int b = b0[z];
		ONECOEFF(r, b, lp0, lp1, lp2);
		b = b1[z];
		ONECOEFF(r, b, lp0, lp1, lp2);
		b = b2[z];
		ONECOEFF(r, b, lp0, lp1, lp2);
		b = b3[z];
		ONECOEFF(r, b, lp0, lp1, lp2);
	}
	
	finish_comp();
}

void LDCTDCTEncoder::encode_color(const u_char* in)
{
	register const short* b0 = (const short*)in;

	register ldct_layer* lp0 = lp_[0];
	register ldct_layer* lp1 = lp_[1];
	register ldct_layer* lp2 = lp_[2];

	int dc = b0[0] >> 3;
	PUT_BITS(dc, 8, lp0->nbb, lp0->bb, lp0->bs);

	lp1->refbs = lp1->ref;
	lp2->refbs = lp2->ref;
	lp1->refnbb = lp2->refnbb = 0;
	lp1->refbb = lp2->refbb = 0;

	u_int r = 0;
	for (int i = 1; i < 64; i++) {
		int z = COLZAG[i];
		register int b = b0[z];
		ONECOEFF(r, b, lp0, lp1, lp2);
	}

	finish_comp();
}

/*
 * XXX
 */
void LDCTEncoder::setup_layers(int i)
{
	int S = (i == 0) ? 1 : 0;
	for (; i < sh_->n; ++i)
		getpkt(&sh_->ls[i], S);
}

#define MAXMBSIZE (16*16+2*8*8+4)

void LDCTEncoder::check_buffers()
{
	for (int i = 0; i < sh_->n; ++i) {
		ldct_layer* lp = &sh_->ls[i];
		if (lp->bs + MAXMBSIZE >= lp->es) {
			flush(i, 0);
			setup_layers(i);
			return;
		}
	}
}

void LDCTEncoder::flush(int i, int sync)
{
	int eog = (i == 0) ? 1 : 0;
	for (; i < sh_->n; ++i)
		flush(&sh_->ls[i], sync, eog);
	if (eog)
		++group_;
}

void LDCTPixelEncoder::recv(Buffer* bp)
{
	const VideoFrame* vf = (VideoFrame*)bp;
	if (!samesize(vf))
		size(vf->width_, vf->height_);
	YuvFrame* p = (YuvFrame*)vf;
#ifdef notdef
	tx_->flush();
#endif
	ts_ = p->ts_;
	cc_ = 0;
	int tl = p->layer_;

	sh_ = &shs_[tl & (NT - 1)];

	u_char* frm = (u_char*)p->bp_;
	const u_char* chm = frm + framesize_;
	blkno_ = -1;
	
	setup_layers(0);

	setup_lps();

	const u_int8_t* crv = p->crvec_;

	int blkw = width_ >> 4;
	int blkh = height_ >> 4;
	int blkno = 0;
	for (int y = 0; y < blkh; ++y) {
		for (int x = 0; x < blkw; ++blkno, frm += 16, chm += 8,
		     ++x, ++crv) {

			int s = crv[0];
			if ((s & CR_SEND) == 0)
				continue;

			check_buffers();
			
			int dblk = blkno - blkno_;
			blkno_ = blkno;
			if (dblk <= 0)
				abort();

			ldct_layer* lp = &sh_->ls[0];

			/* Put addressing information in base layer */

			int nbb = lp->nbb;
			u_int bb = lp->bb;
			u_char* bs = lp->bs;
			if (dblk == 1) {
				PUT_BIT(1, nbb, bb, bs);
			} else if (dblk <= 17) {
				PUT_BITS(0x10 | (dblk - 2), 6, nbb, bb, bs);
			} else {
				PUT_BITS(dblk, 13, nbb, bb, bs);
			}
			lp->nbb = nbb;
			lp->bb = bb;
			lp->bs = bs;
			
			for (int i = 0; i < LDCT_NLAYER; i++) {
				ldct_layer* lp = lp_[i];
				if (lp->mergep != 0) {
					/* Init buffers */
					lp->bs = lp->buf;
					lp->es = lp->bs + mtu_ - HDRSIZE;
					lp->bb = 0;
					lp->nbb = 0;
				}
			}
					
			encode_blk(frm);
			encode_color(chm);
			encode_color(chm + (framesize_ >> 2));
		}
		frm -= width_;
		frm += width_ << 4;
		chm -= width_ >> 1;
		chm += width_ >> 1 << 3;
	}
	
	flush(0, 1);
	nb_ += cc_;
}

/* XXX Merge this with pixel version. */
void LDCTDCTEncoder::recv(Buffer* bp)
{
	const VideoFrame* vf = (VideoFrame*)bp;
	if (!samesize(vf))
		size(vf->width_, vf->height_);
	DCTFrame* p = (DCTFrame*)vf;
#ifdef notdef
	tx_->flush();
#endif
	ts_ = p->ts_;
	cc_ = 0;
	int tl = p->layer_;
	sh_ = &shs_[tl & (NT - 1)];

	short* frm = (short*)p->bp_;
	blkno_ = -1;
	
	setup_layers(0);

	setup_lps();

	const u_int8_t* crv = p->crvec_;

	int blkw = width_ >> 4;
	int blkh = height_ >> 4;
	int blkno = 0;
	for (int y = 0; y < blkh; ++y) {
		for (int x = 0; x < blkw; ++blkno, frm += 6*64, ++x, ++crv) {

			int s = crv[0];
			if ((s & CR_SEND) == 0)
				continue;

			check_buffers();
			
			int dblk = blkno - blkno_;
			blkno_ = blkno;
			if (dblk <= 0)
				abort();

			ldct_layer* lp = &sh_->ls[0];

			/* Put addressing information in base layer */

			int nbb = lp->nbb;
			u_int bb = lp->bb;
			u_char* bs = lp->bs;
			if (dblk == 1) {
				PUT_BIT(1, nbb, bb, bs);
			} else if (dblk <= 17) {
				PUT_BITS(0x10 | (dblk - 2), 6, nbb, bb, bs);
			} else {
				PUT_BITS(dblk, 13, nbb, bb, bs);
			}
			lp->nbb = nbb;
			lp->bb = bb;
			lp->bs = bs;
			
			for (int i = 0; i < LDCT_NLAYER; i++) {
				ldct_layer* lp = lp_[i];
				if (lp->mergep != 0) {
					/* Init buffers */
					lp->bs = lp->buf;
					lp->es = lp->bs + mtu_ - HDRSIZE;
					lp->bb = 0;
					lp->nbb = 0;
				}
			}
					
			encode_blk((u_char*)frm);
			encode_color((u_char*)(frm + 4*64));
			encode_color((u_char*)(frm + 4*64 + 64));
		}
	}
	
	flush(0, 1);
	nb_ += cc_;
}

void LDCTEncoder::setup_lps() 
{
	/* 
	 * Point layer pointers to spatial layer structures.
	 */
	int i;
	for (i = 0; i < sh_->n; ++i) {
		ldct_layer* lp = &sh_->ls[i];
		int bl = LDCT_QD_BL(lp->qd);
		int rlnum = LDCT_QD_RLNUM(lp->qd);

		lp->ht = ldct_hte[bl];
		lp->esc = ldct_esc[bl];
		lp_[bl] = lp;
		/* 
		 * Use temporary storage for "additional"
		 * spatial layers in network layer.
		 * This is clearly suboptimal, but it's a 
		 * reasonable first cut that makes the code 
		 * much simpler at this point.
		 */
		while(rlnum-- > 0) {
			++bl;
			ldct_layer* tlp = &tmpls_[bl];
			
			lp_[bl] = tlp;
			/* 
			 * We point "merge" pointer to the 
			 * ultimate storage destination.
			 * Note that the merge pointer in 
			 * the base layer of the network layer
			 * is always 0 (set at startup).
			 */
			tlp->mergep = lp;
			
		}
	}
}
				
void LDCTEncoder::size(int w, int h)
{
	if ((w | h) & 0xf) {
		printf("vic: bad size in ldct coder (%dx%d)\n", w, h);
		abort();
	}
	FrameModule::size(w, h);
}

void LDCTEncoder::finish_comp()
{
	ldct_layer* lp;

	/* No need to check for merging or refinement in layer 0. */
	lp = lp_[0];
	int val = lp->ht[LDCT_EOB].val;
	register int nb = lp->ht[LDCT_EOB].nb;
	PUT_BITS(val, nb, lp->nbb, lp->bb, lp->bs);

	int i;
	for (i = 1; i < LDCT_NLAYER; i++) {
		lp = lp_[i];

		val = lp->ht[LDCT_EOB].val;
		u_char* ref = lp->ref;
		if (lp->mergep != 0) {
			ldct_layer* mergep = lp->mergep;
			u_char* buf = lp->buf;
			nb = ((lp->bs - buf << 3) + lp->nbb);
			STORE_BITS(lp->bs, lp->bb);
			PUT_BUFFER(mergep->nbb, mergep->bb, mergep->bs, buf, 
				   nb);

			/* Reset data bit buffers */
			lp->bs = lp->buf;
			lp->bb = 0;
			lp->nbb = 0;

			/* XXX */
			nb = lp->ht[LDCT_EOB].nb;
			PUT_BITS(val, nb, mergep->nbb, mergep->bb, mergep->bs);
			register int refnb = ((lp->refbs - ref) << 3) + 
				lp->refnbb;
			if (refnb > 0) {
				STORE_BITS(lp->refbs, lp->refbb);
				PUT_BUFFER(mergep->nbb, mergep->bb, 
					   mergep->bs, ref, refnb);
			}
		} else {
			/* XXX */
			nb = lp->ht[LDCT_EOB].nb;
			PUT_BITS(val, nb, lp->nbb, lp->bb, lp->bs);
			register int refnb = ((lp->refbs - ref) << 3) + 
				lp->refnbb;
			if (refnb > 0) {
				STORE_BITS(lp->refbs, lp->refbb);
				PUT_BUFFER(lp->nbb, lp->bb, lp->bs, ref, 
					   refnb);
			}
		}
	}

}


