#include "mc/request-mc.h"
#include "mc/manager-mc.h"
#include "mc/common-mc.h"

/*
 * Request format
 *  16-bit keylen
 *  16-bit number of RequestPairs
 *  32-bit offset / 32-bit length 
 *  32-bit offset / 32-bit length 
 *  ...
 *  key
 * We could figure out numrps from the packet length,
 * but 16 bits isn't too much to pay for the tradeoff
 * in simplicity. 
 */
//virtual 
int 
MC_Request::fill_ADU(void) {

	Tcl& tcl = Tcl::instance();

	dprintf(stderr, "REQUEST::FILL_ADU():%s\n", key_);

	/* need to get holes from tcl to us */
	int argc; 
	char **argv;
	tcl.evalf("%s get_orequests %s", mcmgr_->name(), key_);		
	Tcl_SplitList(tcl.interp(), tcl.result(), &argc, &argv);
	int offset = atoi(argv[0]);
	if (offset == -1) {
		/* all requests on this url cancelled */
		dprintf(stderr, " ALREADY CANCELLED\n");
		CacheBuf* cb = mcmgr_->lookup_by_key(key_);
		if (cb != 0) {
			cb->request(0);
		}
		this->cancel();
		free(argv);
		return 0;
	}

	/* fill in the RequestPairs */
	Byte *p = pb_->data+sizeof(srmhdr);
	int keylen = strlen(key_);
	int numrps = argc/2;
	*(u_int16_t*)&p[0] = htons(keylen);
	*(u_int16_t*)&p[2] = htons(numrps);
	p += 4;

	int i;
 	for (i = 0; i < numrps; i++) {
		dprintf(stderr, "\tRP: %s %s", argv[2*i], argv[2*i+1]);
//		tcl.evalf("mc_trace_rq %s %s %s", argv[2*i], argv[2*i+1], key_);
		*(u_int32_t*)&p[0]   = htonl(atoi(argv[2*i]));
		*(u_int32_t*)&p[4] = htonl(atoi(argv[2*i+1]));
		p += 8;
	}

	free(argv);

	/* lastly, the key */
	memcpy((char*)p, key_, keylen);

	/* 
	 *  backoff XXX (is this right? do we backoff our own request?) 
	 * or can we see our own traffic?
	 */
	CacheBuf* cb = mcmgr_->lookup_by_key(key_);
	cb->request(0);
	this->backoff();

	/*  XXX check that we didn't exceed SRM_MTU */
	return 4 + argc/2*sizeof(RequestPair) + keylen;
}



