
#
# AnnounceListenManager for CASA clustering client.
# Author: Tina Wong
# Creation Date: April 19, 99
# Status: pre-alpha
#

import AnnounceListenManager/AS/Client

Class AnnounceListenManager/AS/Client/CASA -superclass AnnounceListenManager/AS/Client

#
# Create by application requesting clustering service.
# Parameters are in the form of an key-value list.
#  netspec  = application-level control channel as multicast addr
#  bw       = bandwidth available for channel
#  inst     = unique name to specify application requesting service
# Agent data also as key-value list
#  ctype    = receiver or source in application
#  addrs    = list of blocks of multicast addresses available
#  bw       = total session bandwidth
#  prefops  = specify preference operations script
#  pref     = receiver interest (recv) or data flows (src)
#  alg      = grouping sources or receivers requested
#  layer    = max number of layers on source data (1 means not layerable)
#  partition = default partition to start with
#  adapt    = interval (ms) to recluster
#  thresh   = threshold of mismatch
#
AnnounceListenManager/AS/Client/CASA public init { app params agent_data } {
	$self instvar app_ params_ agent_data_

	set app_ $app
	array set agent_data_ $agent_data
	array set params_ $params
	set srv_loc urn:casa
	$self next $params_(netspec) $params_(bw) $srv_loc
}

AnnounceListenManager/AS/Client/CASA public set_agent_data { l } {
	$self instvar agent_data_

	unset agent_data_
	array set agent_data_ $l
}

AnnounceListenManager/AS/Client/CASA private recv_msg { atype aspec addr srv_name srv_loc srv_inst ssg_port ssd } {
	$self instvar app_

	# pass mapping info from servent to application
	if { "$atype" == "srv" } {
		array set ssd_ $ssd
		$app_ recv_mappings $ssd_(orphans) $ssd_(mappings)
		$app_ recv_bwalloc $ssd_(bwalloc)
	}
}

AnnounceListenManager/AS/Client/CASA public service_name { } {
	return CASA
}

AnnounceListenManager/AS/Client/CASA public service_instance { } {
	$self instvar params_

	return casa-$params_(inst)
}

AnnounceListenManager/AS/Client/CASA public agent_data { } {
	$self instvar agent_data_

	return [array get agent_data_]
}

AnnounceListenManager/AS/Client/CASA public register { atype aspec addr srv_name srv_inst msg } {}

AnnounceListenManager/AS/Client/CASA public unregister { atype aspec addr srv_name srv_inst msg } {}
