
#
# AnnounceListenManager for CASA clustering service.
# Author: Tina Wong
# Creation Date: April 19, 99
# Status: pre-alpha
#

import AnnounceListenManager/AS/Service CASAAdaptTimer
import GS/Kmeans GR/Kmeans
import Prefops/Stockcast Prefops/Newscast

Class AnnounceListenManager/AS/Service/CASA -superclass AnnounceListenManager/AS/Service

#
# Agent data is a key-value list
#  mappings = a list of <source data flows, mcast addr> pairs
#  orphans  = mcast addr for orphan source data flows (those not mapped)
#  bwalloc  = a list of <source data flows, bandwidth> pairs (bw in bits/sec)
#
AnnounceListenManager/AS/Service/CASA public init { argv } {
	array set ssd [lindex $argv 0]
	set srv_inst [lindex $argv 1]
	set megactrl [lindex $argv 3]
	$self next $self $megactrl 20000 $srv_inst

	# clustering related variables
	$self instvar params_
	
	# bw: total session bw avail
	set params_(bw) $ssd(bw)
	
	# A: total number of mcast addr avail
	# addrs: ordered list of mcast addr avail
	# orphans: mcast addr for unmapped sources
	set params_(A) 0
	foreach p $ssd(addrs) {
		set num [lindex $p 1]
		set params_(A) [expr $params_(A) + $num]		
		set start_addr [lindex $p 0]
		lappend addrs $start_addr
		for { set i 1 } { $i < $num } { incr i } {
			lappend addrs [incr_maddr $start_addr $i]
		}
	}
	set params_(A) [expr $params_(A) - 1]
	set params_(addrs) [lrange $addrs 0 [expr $params_(A) - 1]]
	puts stderr "avail addr: $params_(addrs)"
	set params_(orphans) [list [lindex $addrs $params_(A)]]	
	
	# K: total number of clusters used in alg
	set K [expr round($params_(A) / $ssd(layer))]

	# alg: grouping sources or receivers
	set params_(alg) $ssd(alg)

	# thresh: application-dependent threshold on mismatch
	set params_(thresh) $ssd(thresh)
	
	# pref database and operations
	$self instvar prefops_ 
	set prefops_ [new Prefops/$ssd(prefops)]

	# clustering algorithm
	$self instvar alg_
	set alg_ [new $ssd(alg)/Kmeans $prefops_ $K]
	$prefops_ set alg_ $alg_

	# timer for adaptation
	$self instvar timer_
	set timer_ [new CASAAdaptTimer $self $ssd(adapt)]
	$timer_ randomize "yes"
	$timer_ start 
}

proc incr_maddr { start_addr num } {
	set fields [split $start_addr .]
 	set left $num
	for { set i 3 } { $i >= 0 } { incr i -1 } {
		set f [lindex $fields $i]
		if { $i == 0 } {
			set avail [expr 239 - $f]
		} else {
			set avail [expr 255 - $f]
		}
		if { $avail > $left } {
			lappend ans [expr $left + $f]
			for { set j [expr $i - 1]} { $j >= 0 } { incr j -1 } {
				lappend ans [lindex $fields $j]
			}
			break
		} else {
			lappend ans "255"
			set left [expr $left - $avail]
		}
	}
	return "[lindex $ans 3].[lindex $ans 2].[lindex $ans 1].[lindex $ans 0]"
}

AnnounceListenManager/AS/Service/CASA public service_name {} {
	return CASA
}

AnnounceListenManager/AS/Service/CASA public agent_data {} {
	$self instvar params_
	
	set ssd(mappings) [$self get_mappings]
	set ssd(orphans) $params_(orphans)
	set ssd(bwalloc) [$self get_bandwidth]
	return [array get ssd]
}

AnnounceListenManager/AS/Service/CASA public adapt {} {
	$self instvar prefops_ alg_ params_

	# if clustering is not in action yet, do not adapt
	if { [$self get_mappings] == {} } {
		return
	}
	
	$alg_ adapt $params_(thresh)
}

AnnounceListenManager/AS/Service/CASA private get_mappings {} {
	$self instvar alg_ prefops_ params_

	array set cluts [$alg_ get_cluts]
	set num_cluts [$alg_ set num_cluts_]
	if { "$params_(alg)" == "GS" } {
		for {set i 1} {$i <= $num_cluts} {incr i} {
			if [info exist cluts(src,$i)] {
				#puts stderr "cluster #$i: $cluts(src,$i)"
				foreach src_id $cluts(src,$i) {
					set mappings($src_id) [lindex $params_(addrs) [expr $i - 1]]
				}
		        }
		}
		return [array get mappings]
	}
}

AnnounceListenManager/AS/Service/CASA private get_bandwidth {} {
	$self instvar prefops_ params_

	array set l [$prefops_ get_wts]
	foreach src_id [array names l] {
		set bwalloc($src_id) [expr 0.95 * $params_(bw) * $l($src_id)]
	}
	set bwalloc(orphans) [expr $params_(bw) * 0.05]
	return [array get bwalloc]
	
}

AnnounceListenManager/AS/Service/CASA public recv_msg { atype aspec addr srv_name srv_loc srv_inst ssg_port msg } {
	$self instvar prefops_
	
	if { $srv_name == "DEATH" } {
		puts stderr "Received death packet from $aspec at $addr - exiting."
		$self announce_death
		exit 0
	}

	# preference report from receiver
	array set ssd $msg
	if { "$ssd(ctype)" == "recv" } {
		$prefops_ set_pv $addr $ssd(pref)
		
	# heartbeat message from source
	} elseif { "$ssd(ctype)" == "src" } {
		$prefops_ set_df $addr $ssd(pref)
	}
	$self next $atype $aspec $addr $srv_name $srv_loc $srv_inst $ssg_port $msg
}

AnnounceListenManager/AS/Service/CASA instproc register { atype aspec addr srv_name srv_inst msg } {}
		
AnnounceListenManager/AS/Service/CASA instproc unregister { atype aspec addr srv_name srv_inst msg } {
	$self instvar agentbytype_ srv_inst_ agenttab_

	puts stderr "unregister: $atype $aspec $addr $srv_name $srv_inst $msg"

	# If we are alive and a client with a different SID died -- 
	# no need to check.
	if { $atype != "client" || $srv_inst != $srv_inst_ } {
		return
	}

	# notify prefops to age out records
	array set ssd $msg
	
	$self instvar prefops_
	$prefops_ delete $ssd(ctype) $addr 
	
	foreach aspec $agentbytype_(client) {
		set sid [lindex $agenttab_($aspec) 4]
		if { $sid == $srv_inst_ } {
			# Still have a client
			return
		}
	}
	# No clients with our SID, exit
	puts stderr "CASA no more clients -- exiting"
	exit 0
}
