#
# Copyright (c) 1997-1998 Regents of the University of California.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
# 	This product includes software developed by the MASH Research
# 	Group at the University of California Berkeley.
# 4. Neither the name of the University nor of the Research Group may be
#    used to endorse or promote products derived from this software without
#    specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# @(#) $Header: /usr/src/mash/repository/mash/mash-1/tcl/mc/application-mc.tcl,v 1.24 1998/06/28 17:30:11 kwright Exp $
#

import Application SRMAgent AddressBlock MTrace NetworkManager \
       Holes HashTable Launcher 

proc get_mc_drops {} {
	global mc_drops
	return mc_drops
}

proc mc_init_trace trace_name {
	global tracefile
	set tracefile [open $trace_name w]
	after 3000 "mc_auto_flush_trace" 
}

proc mc_auto_flush_trace {} {
	global tracefile
	flush $tracefile
	after 3000 "mc_auto_flush_trace"
}

proc mc_trace_drp {} {
	global tracefile mctime
	set now [$mctime get_time]
	puts $tracefile "$now dp"
}

proc mc_flush_trace {} {
	global tracefile
	flush $tracefile
}

proc mc_close_trace {} {
	global tracefile
	flush $tracefile
	close $tracefile
}

proc mc_trace_ul { url len } {
	# upload to server
	global tracefile mctime
	set now [$mctime get_time]
	puts $tracefile "$now ul $url $len"
}

proc mc_trace_d { offset len url } {
	# data
	global tracefile mctime
	set now [$mctime get_time]
	puts $tracefile "$now d $offset $len $url"
}

proc mc_trace_pf {url pagelen} {
	# page finished
	global tracefile mctime
	set now [$mctime get_time]
	puts $tracefile "$now pf $pagelen $url"
}

proc mc_trace_ps {url pagelen} {
	# page started
	global tracefile mctime
	set now [$mctime get_time]
	puts $tracefile "$now ps $pagelen $url"
}

proc mc_trace_rp { offset len url } {
	# repair
	global tracefile mctime
	set now [$mctime get_time]
	puts $tracefile "$now rp $offset $len $url"
}

proc mc_trace_rq { offset len url } {
	# request
	global tracefile mctime
	set now [$mctime get_time]
	puts $tracefile "$now rq $offset $len $url"
}

proc mc_trace_p { url } {
	# reply-pending
	global tracefile mctime
	set now [$mctime get_time]
	puts $tracefile "$now p $url"
}

Class MCAgent -superclass SRMAgent

MCAgent instproc create-session { appmgr {src_update_handler {}} } {
	$self instvar session_
        set session [new Session/SRM/MC]

        $self app-mgr $appmgr
        $self set src_update_handler_ $src_update_handler
        $session app-mgr $appmgr
        $session agent $self
        $self set session_ $session
	$session_ start_timers
        
        return $session
}

Class WebCache -superclass MCManager


#XXX call back from SRM when src changes
WebCache instproc new_source src {
	puts "new_source $src"
}
WebCache instproc cname_update { src newname } {
	puts "cname_update $src"
}

#
# holes_ missing data in our local cache
# oreplies_ outstanding replies to send
# orequests_ outstanding requests to send
# opendings_ outstanding reply-pendings to send
#
# the above are hash tables whose keys are urls and
# whose elements are Holes objects. For example,
# _oreplies will hole the outstanding replies for
# each url. A Hole object will only be in the hash table
# if it has at least one element.  So if some url
# does not have any outstanding replies, then that
# url will not be in the oreplies_ table. 
#
WebCache public init {sessionAddr sessionPort} {
	$self next
	# associative arrays to map urls to holes; key=url
	$self instvar holes_ oreplies_ orequests_ opendings_
	$self instvar agent_ session_ 
        #set luid [$self resource uid]
        set agent_ [new MCAgent 1]
        set session_ [$agent_ create-session $self $self]
	set ab [new AddressBlock $sessionAddr/$sessionPort]
	$agent_ reset $ab
	delete $ab
	$self init_net
	set holes_ [new HashTable]      
	set oreplies_ [new HashTable]   
	set orequests_ [new HashTable] 
	set opendings_ [new HashTable] 

	# for launching urls
	$self instvar launcher_
	set launcher_ [new Launcher $self]
}

WebCache instproc launch {url httpserver} {
	$self instvar launcher_
	$launcher_ launch $url $httpserver
}
#
# Called from C++ recv() when we have a new page. 
#
WebCache instproc new_page {url pagelen} {
	$self instvar holes_
	$holes_ create $url [new Holes $pagelen]
}

# 
# Called from the proxy server if we
# don't have the entire document. If
# we have no data at all, we register a whole-doc
# request (otherwise we will have a request holes
# structure and a Request object in place). In 
# either case, we register a pending message. 
#
# XXX We should change the multiplier to 0 in case
# we've backed off (check on this). 
#
WebCache instproc request_doc {url server} {
	$self instvar launcher_

	global mc_urlprint
	if {$mc_urlprint == 1} {
		set now [clock seconds]
		puts "$url $now"
	}

	$launcher_ update_proxy_connection $url $server
	if { [$self lookup $url] == "" } {
		$self new_page $url -1
		$self add_request $url 
	} 
	# XXX should only add a pending if we don't have a GET 
	# outstanding
	$self add_pending $url $server
}

#
# Return 1 if we have all the data for this doc
# and 0 otherwise. 
#
WebCache instproc have_all_data {url} {
	$self instvar holes_
	if {[$self lookup $url] != "" 
	    && [$holes_ exists $url] == -1} {
		return yes
	} else {
		return no
	}
}

#
# XXX currently called on every packet!
#
WebCache instproc receive { url off datalen pagelen } {
	$self instvar holes_ orequests_
#puts "RECV $url $off $datalen $pagelen"

	$self cancel_request $url $off $datalen $pagelen
	$self cancel_reply $url $off $datalen $pagelen

	if {[$holes_ exists $url] == -1} {
		# we have all this data and we've cancelled anything
		# we might need to cancel so return. This is special-cased
		# because we can't continue on if we have no holes structure.
		# Usually, this case is due to receiving the data from an
		# http::geturl call. 
		$self check_upload $url $pagelen
		return
	}

	# fill in the data & decide if we need to 
	# sched a request
	set holes [$holes_ get $url]
	# in case we still have an unknown length
	$holes update_length $pagelen 
	set numHoles [$holes fill $off $datalen]
	if {$numHoles > 0} {
		$self add_request $url 
	} else {
		delete $holes
		$holes_ remove $url
		$self check_upload $url $pagelen
	}

}

# Tell the launcher to upload in case there's a proxy connection that
# is waiting for this data. 
WebCache instproc check_upload { url pagelen } {
	if {[$self have_all_data $url] == "yes"} {
		#lookup sets buffer$self. 
		set prefix [$self lookup $url]
		$self instvar launcher_
		$launcher_ upload $url $pagelen $prefix 
                $self cancel_pending $url
#		mc_trace_pf $url $pagelen 
	}

	
}

#
# Called when we receive data
#
WebCache instproc cancel_request { url off len pagelen} {
	$self instvar orequests_
#puts "CANCEL_REQUEST $url $off $len"
	if {[$orequests_ exists $url] == -1} {
		# don't have any outstanding requests
		return
	}
	set oreqs [$orequests_ get $url]
	$oreqs update_length $pagelen
	set numHoles [$oreqs fill $off $len]
	if {$numHoles == 0} {
		delete $oreqs
		$orequests_ remove $url
	}
}

# called when we receive data
WebCache instproc cancel_reply { url off len pagelen} {
	$self instvar oreplies_
#puts "CANCEL_REPLY $url $off $len"
	if {[$oreplies_ exists $url] == -1} {
		# don't have any outstanding requests
		return
	}
	set oreps [$oreplies_ get $url]
	$oreps update_length $pagelen
	set numHoles [$oreps fill $off $len]
	if {$numHoles == 0} {
		delete $oreps
		$oreplies_ remove $url
	}
}

# called when we receive a reply pending msg
WebCache instproc cancel_pending { url } {
	$self instvar opendings_
#puts "CANCEL_PENDING $url"
	if {[$opendings_ exists $url] == -1} {
		# don't have any outstanding pendings
		return
	}
	set opends [$opendings_ get $url]
	delete $opends
	$opendings_ remove $url
}


# Called when a url has been redirected to
# a new url. We need to update all the data
# structures to the new one. 


WebCache instproc update_url { oldurl newurl } {
puts "UPDATE_URL $oldurl -> $newurl"
	$self instvar orequests_ holes_ oreplies_ 
	$holes_ change_key $oldurl $newurl
	$orequests_ change_key $oldurl $newurl
	$oreplies_ change_key $oldurl $newurl
	$self change_key $oldurl $newurl
}

WebCache instproc add_request { url } {
	$self instvar orequests_ holes_
	if {[$orequests_ exists $url] == -1} {
#puts "ADD_REQUEST $url "
		# there are no outstanding requests for this url and we 
		# have some holes. so create a "holes" table, add our holes (via copy)
		# and call send_request. These holes could be a whole-doc request.
		set holes [$holes_ get $url]
		$orequests_ create $url [$holes copy]
		# XXX take the next 2 lines out -- for debug printing only
		set oreqs [$orequests_ get $url]
#		$oreqs print
		$self send_request $url
	}
}

# If there is already a reply pending scheduled,
# don't send a second one. This could happen if
# someone keeps requesting more & more data from a page
# we don't have.
WebCache instproc add_pending { url server } {
	$self instvar opendings_
	global mc_launch
	if { $mc_launch == "no" } {
		# user specified not to launch 
		return
	}
	if {[$opendings_ exists $url] != -1} {
		return
	}
#puts "ADD_PENDING NEW $url $server"
	$opendings_ create $url [new Holes -1]
	set holes [$opendings_ get $url]
	$self send_pending $url $server
}

WebCache instproc add_reply { url offset len } {
	$self instvar oreplies_
#puts "ADD_REPLY: $url $offset $len"
	set newreply 0
	if {[$oreplies_ exists $url] == -1} {
		$oreplies_ create $url [new Holes 4]
		set holes [$oreplies_ get $url]
		$holes fill 0 4
		set newreply 1
	} 
	set holes [$oreplies_ get $url]
	$holes dig $offset $len
	if {$newreply == 1} {
		$self send_reply $url
	}
}

# Return -1 if pending was cancelled.
# If it hasn't been cancelled, delete 
# it and return 1.
WebCache instproc get_pending url {
	$self instvar opendings_
	set opends [$opendings_ exists $url]
	if {$opends == -1} {
		# we're cancelled 
		return -1
	}
	delete $opends
	$opendings_ remove $url
	return 1
}
# Make the holes array of outstanding requests into
# a list and pass it back to the callee; called from
# the MC_Request::fill_ADU; deletes the requests
WebCache instproc get_orequests url {
	$self instvar orequests_
	if {[$orequests_ exists $url] == -1} {
		return -1
	}
	set holes [$orequests_ get $url]
	set answer [$holes listize]
#	delete $holes
#	$orequests_ remove $url
	return $answer
}

# return a single outstanding reply; called from 
# MC_Reply::fill_ADU; delete the reply
WebCache instproc get_oreply url {
	$self instvar oreplies_
	if {[$oreplies_ exists $url] == -1} {
		return -1
	}
	set oreps [$oreplies_ get $url]
	set answer [$oreps remove_first]
	if {$answer != -1} {
		# we do have an unsent reply
		$oreps fill [lindex $answer 0] [lindex $answer 1]
	}
	if {[$oreps numHoles] == 0} {
		delete $oreps
		$oreplies_ remove $url
	}
	return $answer
}

#
# Called from MCManager::handle_request(). Decides
# if we should schedule a reply or a reply pending msg. 
# If we have all the data, schedule a reply. If not,
# schedule a pending. 
# 
# If we have the data, lookup will return something
# and we will not have a holes structure. 
#
# XXX Clean this up
#
WebCache instproc check_reply { url offset len } {
	$self instvar holes_
	set prefix [$self lookup $url]
	set holes [$holes_ exists $url]
#puts "CHECK_REPLY: $url $offset $len "	
	#
	# cover the whole-doc case
	#
	if {$len == -1} {
		if {$prefix != "" && $holes == -1 } {
			# we have the entire document
			set doclen [$self length $url]
			$self add_reply $url $offset $doclen
			return
		} else {
			$self add_pending $url 0
			return
		}		
	}
	#
	# cover the partial doc case
	#
	if {$prefix == ""} {
		# we don't have any data
		$self add_pending $url 0
#puts stderr "TCL:CHECK_REPLY WE DON'T HAVE THIS URL: $url"
		return
	} 
	# by now, we know we have at least some data
	if {$holes == -1} {
		# we have all the data
		$self add_reply $url $offset $len 
		return
	} elseif {[$holes have_data $offset $len] == 1} {
		# we at least have this fragment
		$self add_reply $url $offset $len 
		return
	}
}

WebCache private init_net {} {	
	$self attach_session [$self set session_]
}

WebCache public reset ab {
	#XXX
	$self instvar agent_
	$agent_ net_loopback 1
}

Class MCApp -superclass Application
MCApp instproc init {argv} {
	$self next mc
	$self instvar wc_

	$self add_default defaultTTL 15
        MTrace init trcAll

	set o [$self options]
	$self init_args $o
	$self init_resources $o
	set argv [$o parse_args $argv]
	# argv now has any extra args

	global mc_urlprint
	global mc_drops
	global mc_launch
	global mc_nocache
	set mc_urlprint [$o get_option mcPrintRequests]
	set mc_drops    [$o get_option mcDropPackets]
	set mc_launch   [$o get_option mcLaunch]
	set mc_nocache  [$o get_option mcNoCache]
	set addr        [$o get_option mcSessionAddr]
	set port        [$o get_option mcSessionPort]
	set trace_name  [$o get_option mcTraceFile]
	set seconds     [$o get_option mcExitSeconds]

	# open trace file
	mc_init_trace $trace_name

	# create WebCache
	set wc_ [new WebCache $addr $port]

	# start exit timer
	if { $seconds > 0 } {
		set ms [expr $seconds * 1000]
		after $ms "mc_close_trace; exit 0"
	}

	# begin autopulling 
	set urlfile [$o get_option mcURLList]
	if {$urlfile != "none"} {
		$self auto_pull $urlfile
	}
}


MCApp instproc init_args {o} {
	$o register_option -sa mcSessionAddr
	$o register_option -sp mcSessionPort
	$o register_option -auto mcURLList
	$o register_option -pr mcPrintRequests
	$o register_option -trace mcTraceFile
	$o register_option -exit mcExitSeconds
	$o register_option -drop mcDropPackets
	$o register_option -launch mcLaunch
	$o register_option -nocache mcNoCache
}

MCApp instproc getManager { } {
	return [$self set wc_]
}

MCApp instproc reset { ab } {
	$self instvar wc_
	$wc_ reset $ab
}


MCApp instproc auto_pull_callback { url_index } {
	$self instvar wc_ urls_ url_timings_ 

	if { $url_index < [llength $urls_] - 1} {
		set last [lindex $url_timings_ $url_index]
		set current [lindex $url_timings_ [expr $url_index + 1]]
		set t [expr $current*1000 - $last*1000 + 10]
		after $t "$self auto_pull_callback [expr $url_index + 1]"
	}
	set url [lindex $urls_ $url_index]
	$wc_ request_doc $url 0

}

MCApp instproc auto_pull { urlfile } {
	$self instvar wc_ urls_ url_index_ last_interval_ url_timings_
	set f [open $urlfile]
	while {[gets $f line] >= 0} {
		lappend urls_ [lindex $line 0]
		lappend url_timings_ [lindex $line 1]
	}
	close $f
	# wait a couple seconds to get in sync with other members
	after 3000
	after 0 "$self auto_pull_callback 0"

}

MCApp instproc init_resources {o} {
	$o add_default mcSessionAddr 224.5.5.5 
	$o add_default mcSessionPort 5555
	$o add_default mcURLList none
	$o add_default mcPrintRequests 0
	$o add_default mcTraceFile /dev/null
	$o add_default mcExitSeconds 0
	$o add_default mcDropPackets 0
	$o add_default mcLaunch yes
	$o add_default mcNoCache no
}

